@@ -49,7 +49,8 @@ class AsyncServer(server.Server):
     :param cors_credentials: Whether credentials (cookies, authentication) are
                              allowed in requests to this server.
     :param logger: To enable logging set to ``True`` or pass a logger object to
-                   use. To disable logging set to ``False``.
+                   use. To disable logging set to ``False``. Note that fatal
+                   errors are logged even when ``logger`` is ``False``.
     :param json: An alternative json module to use for encoding and decoding
                  packets. Custom json modules must have ``dumps`` and ``loads``
                  functions that are compatible with the standard library
@@ -199,15 +200,10 @@ class AsyncServer(server.Server):
                 if allowed_origins is not None and origin not in \
                         allowed_origins:
                     self.logger.info(origin + ' is not an accepted origin.')
-                    r = self._bad_request()
-                    make_response = self._async['make_response']
-                    if asyncio.iscoroutinefunction(make_response):
-                        response = await make_response(
-                            r['status'], r['headers'], r['response'], environ)
-                    else:
-                        response = make_response(r['status'], r['headers'],
-                                                 r['response'], environ)
-                    return response
+                    return await self._make_response(
+                        self._bad_request(
+                            origin + ' is not an accepted origin.'),
+                        environ)
 
         method = environ['REQUEST_METHOD']
         query = urllib.parse.parse_qs(environ.get('QUERY_STRING', ''))
@@ -217,6 +213,14 @@ class AsyncServer(server.Server):
         jsonp = False
         jsonp_index = None
 
+        # make sure the client speaks a compatible Engine.IO version
+        sid = query['sid'][0] if 'sid' in query else None
+        if sid is None and query.get('EIO') not in [['2'], ['3']]:
+            return await self._make_response(self._bad_request(
+                'The client is using an unsupported version of the Socket.IO '
+                'or Engine.IO protocols'
+            ), environ)
+
         if 'b64' in query:
             if query['b64'][0] == "1" or query['b64'][0].lower() == "true":
                 b64 = True
@@ -234,7 +238,9 @@ class AsyncServer(server.Server):
         elif method == 'GET':
             if sid is None:
                 transport = query.get('transport', ['polling'])[0]
-                if transport != 'polling' and transport != 'websocket':
+                if (transport != 'polling' and transport != 'websocket') or \
+                        (transport != 'polling'
+                            and transport != environ.get('HTTP_UPGRADE')):
                     self.logger.warning('Invalid transport %s', transport)
                     r = self._bad_request()
                 else:
@@ -294,16 +300,7 @@ class AsyncServer(server.Server):
                         getattr(self, '_' + encoding)(r['response'])
                     r['headers'] += [('Content-Encoding', encoding)]
                     break
-        cors_headers = self._cors_headers(environ)
-        make_response = self._async['make_response']
-        if asyncio.iscoroutinefunction(make_response):
-            response = await make_response(r['status'],
-                                           r['headers'] + cors_headers,
-                                           r['response'], environ)
-        else:
-            response = make_response(r['status'], r['headers'] + cors_headers,
-                                     r['response'], environ)
-        return response
+        return await self._make_response(r, environ)
 
     def start_background_task(self, target, *args, **kwargs):
         """Start a background task using the appropriate async model.
@@ -362,6 +359,21 @@ class AsyncServer(server.Server):
         """
         return asyncio.Event(*args, **kwargs)
 
+    async def _make_response(self, response_dict, environ):
+        cors_headers = self._cors_headers(environ)
+        make_response = self._async['make_response']
+        if asyncio.iscoroutinefunction(make_response):
+            response = await make_response(
+                response_dict['status'],
+                response_dict['headers'] + cors_headers,
+                response_dict['response'], environ)
+        else:
+            response = make_response(
+                response_dict['status'],
+                response_dict['headers'] + cors_headers,
+                response_dict['response'], environ)
+        return response
+
     async def _handle_connect(self, environ, transport, b64=False,
                               jsonp_index=None):
         """Handle a client connection request."""
@@ -383,14 +395,14 @@ class AsyncServer(server.Server):
 
         ret = await self._trigger_event('connect', sid, environ,
                                         run_async=False)
-        if ret is False:
+        if ret is not None and ret is not True:
             del self.sockets[sid]
             self.logger.warning('Application rejected connection')
-            return self._unauthorized()
+            return self._unauthorized(ret or None)
 
         if transport == 'websocket':
             ret = await s.handle_get_request(environ)
-            if s.closed:
+            if s.closed and sid in self.sockets:
                 # websocket connection ended, so we are done
                 del self.sockets[sid]
             return ret
@@ -398,7 +410,18 @@ class AsyncServer(server.Server):
             s.connected = True
             headers = None
             if self.cookie:
-                headers = [('Set-Cookie', self.cookie + '=' + sid)]
+                if isinstance(self.cookie, dict):
+                    headers = [(
+                        'Set-Cookie',
+                        self._generate_sid_cookie(sid, self.cookie)
+                    )]
+                else:
+                    headers = [(
+                        'Set-Cookie',
+                        self._generate_sid_cookie(sid, {
+                            'name': self.cookie, 'path': '/', 'SameSite': 'Lax'
+                        })
+                    )]
             try:
                 return self._ok(await s.poll(), headers=headers, b64=b64,
                                 jsonp_index=jsonp_index)
@@ -459,7 +482,12 @@ class AsyncServer(server.Server):
                     if not socket.closing and not socket.closed:
                         await socket.check_ping_timeout()
                     await self.sleep(sleep_interval)
-            except (SystemExit, KeyboardInterrupt, asyncio.CancelledError):
+            except (
+                SystemExit,
+                KeyboardInterrupt,
+                asyncio.CancelledError,
+                GeneratorExit,
+            ):
                 self.logger.info('service task canceled')
                 break
             except:
