@@ -48,9 +48,14 @@ class Server(object):
     :param compression_threshold: Only compress messages when their byte size
                                   is greater than this value. The default is
                                   1024 bytes.
-    :param cookie: Name of the HTTP cookie that contains the client session
-                   id. If set to ``None``, a cookie is not sent to the client.
-                   The default is ``'io'``.
+    :param cookie: If set to a string, it is The name of the HTTP cookie the
+                   server sends back tot he client containing the client
+                   session id. If set to a dictionary, the ``'name'`` key
+                   contains the cookie name and other keys define cookie
+                   attributes, where the value of each attribute can be a
+                   string, a callable with no arguments, or a boolean. If set
+                   to ``None``, a cookie is not sent to the client. The
+                   default is ``'io'``.
     :param cors_allowed_origins: Origin or list of origins that are allowed to
                                  connect to this server. Only the same origin
                                  is allowed by default. Set this argument to
@@ -331,18 +336,26 @@ class Server(object):
                 if allowed_origins is not None and origin not in \
                         allowed_origins:
                     self.logger.info(origin + ' is not an accepted origin.')
-                    r = self._bad_request()
+                    r = self._bad_request(
+                        origin + ' is not an accepted origin.')
                     start_response(r['status'], r['headers'])
                     return [r['response']]
 
         method = environ['REQUEST_METHOD']
         query = urllib.parse.parse_qs(environ.get('QUERY_STRING', ''))
-
-        sid = query['sid'][0] if 'sid' in query else None
         b64 = False
         jsonp = False
         jsonp_index = None
 
+        # make sure the client speaks a compatible Engine.IO version
+        sid = query['sid'][0] if 'sid' in query else None
+        if sid is None and query.get('EIO') not in [['2'], ['3']]:
+            r = self._bad_request(
+                'The client is using an unsupported version of the Socket.IO '
+                'or Engine.IO protocols')
+            start_response(r['status'], r['headers'])
+            return [r['response']]
+
         if 'b64' in query:
             if query['b64'][0] == "1" or query['b64'][0].lower() == "true":
                 b64 = True
@@ -360,7 +373,9 @@ class Server(object):
         elif method == 'GET':
             if sid is None:
                 transport = query.get('transport', ['polling'])[0]
-                if transport != 'polling' and transport != 'websocket':
+                if (transport != 'polling' and transport != 'websocket') or \
+                        (transport != 'polling'
+                            and transport != environ.get('HTTP_UPGRADE')):
                     self.logger.warning('Invalid transport %s', transport)
                     r = self._bad_request()
                 else:
@@ -486,6 +501,20 @@ class Server(object):
         """Generate a unique session id."""
         return uuid.uuid4().hex
 
+    def _generate_sid_cookie(self, sid, attributes):
+        """Generate the sid cookie."""
+        cookie = attributes.get('name', 'io') + '=' + sid
+        for attribute, value in attributes.items():
+            if attribute == 'name':
+                continue
+            if callable(value):
+                value = value()
+            if value is True:
+                cookie += '; ' + attribute
+            else:
+                cookie += '; ' + attribute + '=' + value
+        return cookie
+
     def _handle_connect(self, environ, start_response, transport, b64=False,
                         jsonp_index=None):
         """Handle a client connection request."""
@@ -506,14 +535,14 @@ class Server(object):
         s.send(pkt)
 
         ret = self._trigger_event('connect', sid, environ, run_async=False)
-        if ret is False:
+        if ret is not None and ret is not True:
             del self.sockets[sid]
             self.logger.warning('Application rejected connection')
-            return self._unauthorized()
+            return self._unauthorized(ret or None)
 
         if transport == 'websocket':
             ret = s.handle_get_request(environ, start_response)
-            if s.closed:
+            if s.closed and sid in self.sockets:
                 # websocket connection ended, so we are done
                 del self.sockets[sid]
             return ret
@@ -521,7 +550,18 @@ class Server(object):
             s.connected = True
             headers = None
             if self.cookie:
-                headers = [('Set-Cookie', self.cookie + '=' + sid)]
+                if isinstance(self.cookie, dict):
+                    headers = [(
+                        'Set-Cookie',
+                        self._generate_sid_cookie(sid, self.cookie)
+                    )]
+                else:
+                    headers = [(
+                        'Set-Cookie',
+                        self._generate_sid_cookie(sid, {
+                            'name': self.cookie, 'path': '/', 'SameSite': 'Lax'
+                        })
+                    )]
             try:
                 return self._ok(s.poll(), headers=headers, b64=b64,
                                 jsonp_index=jsonp_index)
@@ -580,11 +620,14 @@ class Server(object):
                     'headers': [('Content-Type', 'text/plain')],
                     'response': b'OK'}
 
-    def _bad_request(self):
+    def _bad_request(self, message=None):
         """Generate a bad request HTTP error response."""
+        if message is None:
+            message = 'Bad Request'
+        message = packet.Packet.json.dumps(message)
         return {'status': '400 BAD REQUEST',
                 'headers': [('Content-Type', 'text/plain')],
-                'response': b'Bad Request'}
+                'response': message.encode('utf-8')}
 
     def _method_not_found(self):
         """Generate a method not found HTTP error response."""
@@ -592,11 +635,14 @@ class Server(object):
                 'headers': [('Content-Type', 'text/plain')],
                 'response': b'Method Not Found'}
 
-    def _unauthorized(self):
+    def _unauthorized(self, message=None):
         """Generate a unauthorized HTTP error response."""
+        if message is None:
+            message = 'Unauthorized'
+        message = packet.Packet.json.dumps(message)
         return {'status': '401 UNAUTHORIZED',
-                'headers': [('Content-Type', 'text/plain')],
-                'response': b'Unauthorized'}
+                'headers': [('Content-Type', 'application/json')],
+                'response': message.encode('utf-8')}
 
     def _cors_allowed_origins(self, environ):
         default_origins = []
